from sage.structure.unique_representation import UniqueRepresentation 
from sage.structure.parent import Parent 
from sage.structure.element import RingElement

from sage.rings.infinity import Infinity 
from sage.rings.integer_ring import ZZ


class ParentNewtonPolygon(Parent,UniqueRepresentation):
    def __repr__(self):
        return "Parent for Newton polygons"

ParentNP = ParentNewtonPolygon()


class NewtonPolygon_lastslope(RingElement):
    def __init__(self,vertices,lastslope=Infinity,check=True):
        RingElement.__init__(self, ParentNP)
        if not isinstance(vertices,list):
            raise TypeError("vertices must be a list of coordinates (x,y)")
        if lastslope < Infinity and len(vertices) == 0:
            raise ValueError("vertices can't be empty when lastslope is not Infinity")
        for p in vertices:
            if not isinstance(p,tuple) or len(p) != 2:
                raise TypeError("vertices must be a list of coordinates (x,y)")
        self._vertices = vertices
        self._lastslope = lastslope
        if check: self._normalize()

    def _normalize(self):
        # compute convex hull complexity: O(n*log(n))
        vertices = { }
        minx = Infinity
        for (x,y) in self._vertices:
            if vertices.has_key(x):
                if y < vertices[x]: vertices[x] = y
            else:
                if y < Infinity:
                    vertices[x] = y
                    if x < minx: minx = x
        if minx is Infinity:
            self._vertices = [ ]
            return
        miny = vertices[minx]
        convexhull = [ (minx, miny) ]
        del vertices[minx]
        abscissas = { }
        for x in vertices:
            slope = (vertices[x]-miny) / (x-minx)
            if not abscissas.has_key(slope) or abscissas[slope] < x:
                abscissas[slope] = x
        slopes = abscissas.keys()
        slopes.sort()
        for slope in slopes:
            if abscissas[slope] > minx:
                x = abscissas[slope]
                y = vertices[x]
                if (miny-y)/(minx-x) >= self._lastslope:
                    break
                convexhull.append((x,y))
                (minx, miny) = (x,y)
        self._vertices = convexhull

    def __repr__(self):
        length = len(self._vertices)
        if self._lastslope is Infinity:
            if length == 0:
                return "Empty Newton Polygon"
            elif length == 1:
                return "Finite Newton polygon with 1 vertex: %s" % self._vertices[0]
            else:
                return "Finite Newton polygon with %s vertices: %s" % (length, str(self._vertices)[1:-1])
        else:
            if length == 1:
                return "Newton Polygon consisting of a unique infinite line of slope %s starting at %s" % (self._lastslope, self._vertices[0])
            else:
                return "Infinite Newton polygon with %s vertices: %s ending by an infinite line of slope %s" % (length, str(self._vertices)[1:-1], self._lastslope)

    def vertices(self):
        return self._vertices

    def last_slope(self):
        return self._lastslope

    def slopes(self):
        slopes = [ ]
        for i in range(1,len(self._vertices)):
            dx = self._vertices[i][0] - self._vertices[i-1][0]
            dy = self._vertices[i][1] - self._vertices[i-1][1]
            slope = dy/dx
            slopes.extend(dx * [slope])
        return slopes

    def _add_(self,other):
        return self.__class__(self._vertices + other._vertices, min(self._lastslope, other._lastslope))

    def _mul_(self,other):
        x = y = ZZ(0)
        iself = iother = 0
        slopeself = slopeother = -Infinity
        (dxself, dyself) = self._vertices[0]
        (dxother, dyother) = other._vertices[0]
        vertices = [ ]
        lastslope = min(self._lastslope, other._lastslope)
        while slopeself < lastslope or slopeother < lastslope:
            c = cmp(slopeself, slopeother)
            if c <= 0:
                x += dxself
                y += dyself
                iself += 1
                if iself < len(self._vertices):
                    dxself = self._vertices[iself][0] - self._vertices[iself-1][0]
                    dyself = self._vertices[iself][1] - self._vertices[iself-1][1]
                    slopeself = dyself/dxself
                else:
                    slopeself = lastslope
            if c >= 0:
                x += dxother
                y += dyother
                iother += 1
                if iother < len(other._vertices):
                    dxother = other._vertices[iother][0] - other._vertices[iother-1][0]
                    dyother = other._vertices[iother][1] - other._vertices[iother-1][1]
                    slopeother = dyother/dxother
                else:
                    slopeother = lastslope
            vertices.append((x,y))
        return self.__class__(vertices, lastslope, check=False)

    def __pow__(self,exp,ignored=None):
        vertices = [ (exp*x, exp*y) for (x,y) in self._vertices ]
        return self.__class__(vertices, self._lastslope, check=False)

    def __lshift__(self,i):
        vertices = [ (x, y+i) for (x,y) in self._vertices ]
        return self.__class__(vertices, self._lastslope, check=False)
    def __rshift__(self,i):
        vertices = [ (x, y-i) for (x,y) in self._vertices ]
        return self.__class__(vertices, self._lastslope, check=False)

    def __call__(self,x):
        # complexity: O(log(n))
        from sage.functions.other import floor
        vertices = self._vertices
        if len(vertices) == 0 or x < vertices[0][0]:
            return Infinity
        if x == vertices[0][0]:
            return vertices[0][1]
        if x == vertices[-1][0]:
            return vertices[-1][1]
        if x > vertices[-1][0]:
            return vertices[-1][1] + self._lastslope * (x - vertices[-1][0])
        a = 0; b = len(vertices)
        while b - a > 1:
            c = floor((a+b)/2)
            if vertices[c][0] < x:
                a = c
            else:
                b = c
        (xg,yg) = vertices[a]
        (xd,yd) = vertices[b]
        return ((x-xg)*yd + (xd-x)*yg) / (xd-xg)

    def __eq__(self,other):
        # We assume that other is also an object NewtonPolygon_lastslope
        return self._lastslope == other._lastslope and self._vertices == other._vertices

    def __ne__(self,other):
        return not (self == other)

    def __le__(self,other):
        # We assume that other is also an object NewtonPolygon_lastslope
        if self._vertices[0][0] > other._vertices[0][0]:
            return False
        if self._lastslope > other._lastslope:
            return False

        # Note: a better algorithm exists
        for (x,y) in self._vertices:
            if other(x) < y:
                return False
        return True

    def __lt__(self,other):
        return self <= other and self != other

    def __ge__(self,other):
        return other <= self

    def __gt__(self,other):
        return other <= self and self != other
        
    def plot(self,**kwargs):
        vertices = self._vertices
        if len(vertices) == 0:
            from sage.plot.graphics import Graphics
            return Graphics()
        else:
            from sage.plot.line import line
            (xstart,ystart) = vertices[0]
            (xend,yend) = vertices[-1]
            if self._lastslope is Infinity:
                return line([(xstart, ystart+1), (xstart,ystart+0.5)], linestyle="--", **kwargs) \
                     + line([(xstart, ystart+0.5)] + vertices + [(xend, yend+0.5)], **kwargs) \
                     + line([(xend, yend+0.5), (xend, yend+1)], linestyle="--", **kwargs)
            else:
                return line([(xstart, ystart+1), (xstart,ystart+0.5)], linestyle="--", **kwargs) \
                     + line([(xstart, ystart+0.5)] + vertices + [(xend+0.5, yend + 0.5*self._lastslope)], **kwargs) \
                     + line([(xend+0.5, yend + 0.5*self._lastslope), (xend+1, yend+self._lastslope)], linestyle="--", **kwargs)

